{*****************************************************************************}
{                                                                             }
{  ESSHtmlReport Component                                                    }
{                                                                             }
{  Copyright(C) 2002                                                          }
{  Ecological Software Solutions  (ESS)                                       }
{  All rights reserved.                                                       }
{                                                                             }
{                                                                             }
{  This unit may be used free of charge in commercial and non-commercial      }
{  application provided all the following conditions are followed.            }
{                                                                             }
{  Ecological Software Solutions does not guarrantee this unit in any way.    }
{  Use at your own risk.                                                      }
{                                                                             }
{  The user of this component may alter this code but any changes             }
{  to this code must still retain this copyright notice.  Alterations of this }
{  unit are derived works under US and international copyright laws and can   }
{  not be copyrighted as orginal works or distributed in any way.  Ecological }
{  Software Solutions retains all copyrights to this unit and any derived     }
{  works based on this unit.  The user of this component may not claim any    }
{  copyright on this unit, or distribute this unit or a derived work based on }
{  this unit.                                                                 }
{                                                                             }
{  The user of this component may use inheritance to create separate          }
{  components based on this unit.  The user may copyright and distribute an   }
{  inherited unit as an original work, but the user may only copyright and    }
{  distribute the registered user's component's units and classes. The user   }
{  has at no time the right to distribute this unit with the registered       }
{  user's components.  For another party to use the user's component(s)       }
{  based on this unit, the other party must acquire this unit separately      }
{  from a source authorized by Ecological Software Solutions.                 }
{                                                                             }
{  Any user only has permission to compile this code, or any code created     }
{  from altering this code (including but not exclusive to derived            }
{  components) into a machine readable only executable program.  This unit    }
{  in part or whole can only be compiled into a library (including a DLL,     }
{  OCX, delphi package(s) that is accessible only from the software you       }
{  develope.  All other library uses are not permitted without prior written  }
{  permission from Ecological Software Solutions.                             }
{                                                                             }
{  It is requested that alterations or corrections to this unit be emailed to }
{  Ecological Software Solutions.  This, of course, does not apply to         }
{  inherited component classes.                                               }
{                                                                             }
{  If you find this component useful, please let us know as it helps us       }
{  decide if we should design other, simlar products.                         }
{                                                                             }
{  Contact by email at: mail@ecostats.com                                     }
{                                                                             }
{*****************************************************************************}

unit ESSHtmlReport;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  HTTPApp, DBWeb, DsgnIntf, db;

type

  TOperand = object
    ParIndex   : integer; {if not negative, it points to elem. of Parameters}
    ActualVal : Variant; {actual numeric value if ParIndex < 0}
    function Value : Variant;
  end;

  THtmlOnEvaluateFunction = procedure(const x, y: variant; const FunctionName: String; var FuncResult: Variant) of Object;

  THtmlReportPostfixItem = object
    op : Word; //char;          {operation code or nop}
    Operand : TOperand; {valid only if op = nop}
  end;

  THtmlReportPostfix = object
    Count : integer;
    Items : array of THtmlReportPostfixItem;
  end;

  THtmlReportFunction = class
  private
    FIsCorrect, FFilterBoolean : boolean;
    FCount, FGroupRecordEnd, FGroupRecordStart : Integer;
    FParamsList, TempPfxLine : TStringList;
    FWorkStr, FToken, FFilterStr : string;
    cur, FWorkStrLength : integer;
    FEvaluateFunction : THtmlOnEvaluateFunction;
    FFunctions   : TStringList;
    FGrouped : Boolean;
    FDataSet : TDataSet;
    PostfixItems : array of THtmlReportPostfix;
    procedure OmitComment(var CurLine : String);
    procedure ConvertOperators(var CurLine : String);
    procedure myTrim(var s : string);
    function  CurSegm : string;
    function  ParseLine(const CurLine: String): Boolean;
    procedure SetFunctions(FunctionList : TStringList);
    procedure CheckIfFinished;
    procedure NextToken;
    procedure Expression;
    procedure Init;
    function  FilterOperations(op: Word; y: Variant; x: Variant) : Variant;
    function  PostfixEvaluate(i: Integer) : Variant;
    procedure SetFilterString(const ValString : String);
    constructor create;
    destructor Destroy; override;
    property Count : integer read FCount write FCount;
    property EvaluationFunction : String read FFilterStr write SetFilterString;
    function Evaluate : Variant;
    property Grouped : Boolean read FGrouped write FGrouped;
    property GroupRecordEnd : Integer read FGroupRecordEnd write FGroupRecordEnd;
    property GroupRecordStart : Integer read FGroupRecordStart write FGroupRecordStart;
    property IsCorrect : boolean read FIsCorrect;
    property DataSet : TDataSet read FDataset write FDataSet;
  public
    property Functions : TStringList read FFunctions write SetFunctions;
    property OnEvaluateFunction : THtmlOnEvaluateFunction read FEvaluateFunction write FEvaluateFunction;
  end;

  TESSHtmlReport = class;

  TEssHtmlReportGroups = Class(TPersistent)
  private
    FEssHtmlReport : TESSHtmlReport;
    FGrouped : Boolean;
    FGroupedFields : String;
    FGroupedFieldsColor : THTMLBgColor;
    FGroupFieldsList : TStringList;
    FGroupListValues : TStringList;
    procedure ParseFieldString(AFields: String; ResultList: TStringList);
    procedure SetGrouped(IsGrouped: Boolean);
    procedure SetGroupedFields(AGroupedFields: String);
    procedure SetGroupedFieldsColor(AColor: THTMLBgColor);
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor create(ESSHtmlReport: TESSHtmlReport);
    destructor Destroy;
  published
    property GroupedFields : String read FGroupedFields write SetGroupedFields;
    property Grouped : Boolean read FGrouped write SetGrouped default False;
    property GroupedFieldsColor : THTMLBgColor read FGroupedFieldsColor write SetGroupedFieldsColor;
  end;

  TESSHtmlReport = class(TDataSetTableProducer)
  private
    { Private declarations }
    FDataCellColor : THTMLBgColor;
    FGroups : TEssHtmlReportGroups;
    FEmptyCellColor : THTMLBgColor;
    FFontFace : String;
    FFontSize : Integer;
    FShowNiceFields : Boolean;
    FPageText : TStringList;
    FReportFunctions : THtmlReportFunction;
    FReportSource : String;
    FReportTitle : String;
    FSummaryFields : String;
    FSummaryFieldsList : TStringList;
    FSummaryFunctionList : TStringList;
    FEvaluateFunction : THtmlOnEvaluateFunction;
    function GetFunctions(APageString: String): String;
    function GetHtmlText: String;
    function GetSQL: String;
    function GetSummaryRow: String;
    function GetSummaryFields(APageString: String): String;
    procedure ParseFieldString(AFields: String; ResultList: TStringList);
    procedure SetGroups(AGroups: TEssHtmlReportGroups);
    procedure SetHtmlText(SText: String);
    procedure SetReportSource(const AFileName: String);
    procedure SetSummaryFields(const ASummaryFields: String);
  protected
    { Protected declarations }
    procedure SetDataSet(ADataSet: TDataSet); override;
    procedure SetColumns; dynamic;
    procedure SetColumnFormat(TableString: String);
    procedure SetGroupsFromDocument;
    procedure SetTableFormat(TableString: String);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Content: String; override;
    property ReportFunctions : THtmlReportFunction read FReportFunctions;
    property HtmlText : String Read GetHtmlText write SetHtmlText;
    property PageText : TStringList read FPageText;
    property SQL : String read GetSQL; //report defined SQL on Html page
    procedure DoFormatCell(CellRow, CellColumn: Integer; var BgColor: THTMLBgColor;
        var Align: THTMLAlign; var VAlign: THTMLVAlign;
        var CustomAttrs, CellData: string); override;
  published
    { Published declarations }
    property DataCellColor : THTMLBgColor Read FDataCellColor write FDataCellColor;
    property Groups : TEssHtmlReportGroups read FGroups write SetGroups;
    property EmptyCellColor : THTMLBgColor Read FEmptyCellColor write FEmptyCellColor;
    property FontFace : String read FFontFace write FFontFace;
    property FontSize : Integer read FFontSize write FFontSize;
    property ShowNiceFields : Boolean read FShowNiceFields write FShowNiceFields;
    property SummaryFields : String read FSummaryFields write SetSummaryFields;
    property ReportSource : String read FReportSource write SetReportSource;
    property ReportTitle : String read FReportTitle write FReportTitle;
    property OnEvaluateFunction : THtmlOnEvaluateFunction read FEvaluateFunction write FEvaluateFunction;
  end;

// procedure Register; {Only enable this if you must compile without the property editor}

implementation

uses Math;

const AlfaNum = ['.', ',', ' ', '"', '_', '0'..'9', 'A'..'Z', 'a'..'z']; // include the ',' for european decimal notation
      AllowedOp = '@%*+-/^=<>&|'; //&:AND, |:OR, @:>=, %:<=, :<>
      nop = 0;
var
     Parameters : array of Variant;


{Only enable the register procedure if you must compile without the property editor}
{procedure Register;
begin
  RegisterComponents('ESS', [TESSHtmlReport]);
  RegisterPropertyEditor(TypeInfo(string), TESSHtmlReport, 'ReportSource', TReportSourceFileName);
end;}

// Grouping class //

constructor TEssHtmlReportGroups.Create(ESSHtmlReport: TESSHtmlReport);
begin
  inherited Create;
  FEssHtmlReport := ESSHtmlReport;
  FGroupFieldsList := TStringList.Create;
  FGroupListValues := TStringList.Create;
end;

destructor TEssHtmlReportGroups.Destroy;
begin
  FGroupFieldsList.Free;
  FGroupListValues.Free;
  inherited Destroy;
end;

procedure TEssHtmlReportGroups.AssignTo(Dest: TPersistent);
begin
  IF Dest IS TEssHtmlReportGroups THEN
    WITH TEssHtmlReportGroups(Dest) DO
      BEGIN
        GroupedFields := Self.FGroupedFields;
        GroupedFieldsColor := Self.FGroupedFieldsColor;
        Grouped := Self.FGrouped;
      END
    ELSE
     inherited AssignTo(Dest);
end;

procedure TEssHtmlReportGroups.SetGrouped(IsGrouped: Boolean);
begin
  IF (csDesigning IN FEssHtmlReport.ComponentState) AND (FGroupedFields = '') AND IsGrouped THEN
    BEGIN
      ShowMessage('Please first asign grouped fields.');
      FEssHtmlReport.FReportFunctions.Grouped := False;
    END
  ELSE
    FGrouped := IsGrouped;
  FEssHtmlReport.FReportFunctions.Grouped := IsGrouped AND (GroupedFields<>'');
end;

procedure TEssHtmlReportGroups.SetGroupedFields(AGroupedFields: String);
begin
  FGroupedFields := AGroupedFields;
  IF FGroupedFields <> '' THEN
    ParseFieldString(FGroupedFields, FGroupFieldsList)
  ELSE BEGIN
    FGrouped := False;
    FGroupFieldsList.Clear;
  END;
  FEssHtmlReport.FReportFunctions.Grouped := FGrouped AND (AGroupedFields<>'');
end;

procedure TEssHtmlReportGroups.SetGroupedFieldsColor(AColor: THTMLBgColor);
begin
  FGroupedFieldsColor := AColor;
end;

procedure TEssHtmlReportGroups.ParseFieldString(AFields: String; ResultList: TStringList);
var
   StrSepLoc : Integer;
begin
  ResultList.Clear;
  IF AFields = '' THEN Exit;
  Repeat
    StrSepLoc := AnsiPos(';', AFields); // Find the location of the next separator
    IF StrSepLoc = 0 THEN
      ResultList.Add(AFields)
    ELSE BEGIN
      ResultList.Add(Copy(AFields, 1, StrSepLoc - 1));
      Delete(AFields, 1, StrSepLoc);
    END;
    Trim(AFields);
  Until StrSepLoc = 0;
end;

// Reporting class //

constructor TEssHtmlReport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReportFunctions := THtmlReportFunction.Create;
  OnEvaluateFunction := FReportFunctions.OnEvaluateFunction;
  FSummaryFieldsList := TStringList.Create;
  FSummaryFunctionList := TStringList.Create;
  FGroups := TEssHtmlReportGroups.Create(Self);
  FShowNiceFields := True;
  FFontFace := 'Default';
  FFontSize := 2;
  MaxRows := -1; // No row limit by default
  FPageText := TStringList.Create;
  // Add the mosts commonly used functions as available by default
  FReportFunctions.Functions.Add('count');
  FReportFunctions.Functions.Add('sum');
  FReportFunctions.Functions.Add('avg');
  FReportFunctions.Functions.Add('max');
  FReportFunctions.Functions.Add('min');
end;

destructor TEssHtmlReport.Destroy;
begin
  FSummaryFieldsList.Free;
  FSummaryFunctionList.Free;
  FPageText.Free;
  FReportFunctions.Free;
  inherited Destroy;
end;

function TEssHtmlReport.Content: String;
var
  EmbeddedReport: Boolean;
  i, TableIndex : Integer;
  TempHeader, TempFooter : String;
  APageString, ATableString, AFieldValue : String;
begin
  IF FPageText.Count = 0 THEN BEGIN
    EmbeddedReport := True;
    FPageText.AddStrings(Header);
    FPageText.AddStrings(Footer);
    TempHeader := Header.Text;
    TempFooter := Footer.Text;
    Header.Clear;
    Footer.Clear;
    IF Pos('%add_summary%', lowercase(TempFooter)) > 0 THEN
      Footer.Add('%add_summary%');
    SetColumns;
    SetGroupsFromDocument;
    FPageText.Text := StringReplace(FPageText.Text, '%add_summary%', '', [rfReplaceAll, rfIgnoreCase]);
  END;
  APageString := FPageText.Text;
  IF Assigned(DataSet) THEN BEGIN
    TableIndex := Pos('%insert_table%', APageString);
    IF TableIndex >= 1 THEN BEGIN
      ATableString := inherited Content;
      IF EmbeddedReport THEN BEGIN
        Header.Text := TempHeader;
        Footer.Text := TempFooter;
      END;
      APageString := StringReplace(APageString, '%insert_table%', ATableString, [rfIgnoreCase]);
    END;
    FOR i := 0 TO DataSet.Fields.Count - 1 DO BEGIN
      IF Trim(DataSet.Fields[i].AsString) = '' THEN
        AFieldValue := '&nbsp'
      ELSE
        AFieldValue := DataSet.Fields[i].AsString;
      APageString := StringReplace(APageString, '%field='+DataSet.Fields[i].FieldName+'%', AFieldValue, [rfReplaceAll, rfIgnoreCase]);
    END;
    IF (FSummaryFieldsList.Count > 0) THEN
      APageString := GetSummaryFields(APageString);
    APageString := GetFunctions(APageString);
  END;

  APageString := StringReplace(APageString, '%table_header', FReportTitle, [rfReplaceAll, rfIgnoreCase]);
  APageString := StringReplace(APageString, '%table_footer', FReportTitle, [rfReplaceAll, rfIgnoreCase]);
  APageString := StringReplace(APageString, '%report_title%', FReportTitle, [rfReplaceAll, rfIgnoreCase]);
  FPageText.Text := StringReplace(APageString, '%report_date%', DateToStr(Now), [rfReplaceAll, rfIgnoreCase]);
end;

procedure TEssHtmlReport.SetDataSet(ADataSet: TDataSet);
begin
  inherited SetDataSet(ADataSet);
  FReportFunctions.DataSet := ADataSet;
end;

procedure TEssHtmlReport.SetColumns;
var
  ColumnIndex, TableIndex, TableEndIndex : Integer;
  PageString, TableString : String;
begin
  IF FPageText = Nil THEN
    ShowMessage('Report source not assigned.')
  ELSE BEGIN
    PageString := FPageText.Text;
    ColumnIndex := Pos('%use_these_fields%', PageString);
    IF ColumnIndex > 0 THEN BEGIN
      TableString := Trim(Copy(PageString, ColumnIndex + Length('%use_these_fields%'), Length(PageString) - ColumnIndex));
      TableEndIndex := Pos('</table>', TableString);
      TableIndex := Pos('<table', TableString);
      TableString := Trim(Copy(TableString, TableIndex, Length(TableString) - TableIndex - 1));
      TableIndex := Pos('</table>', TableString);
      TableString := Trim(Copy(TableString, 1, TableIndex + Length('</table>')));
      SetTableFormat(TableString);
      SetColumnFormat(TableString);
      Delete(PageString, ColumnIndex, TableEndIndex + 25);
      FPageText.Text := PageString;
    END;
  END;
end;

procedure TEssHtmlReport.SetGroupsFromDocument;
var
  GroupPos : Integer;
  TempStr : String;
begin
  TempStr := LowerCase(FPageText.Text);
  GroupPos := Pos('%grouped_fields=', TempStr);
  IF GroupPos > 0 THEN BEGIN
    Delete(TempStr, 1, GroupPos + Length('%grouped_fields=') - 1);
    GroupPos := Pos('%', TempStr);
    Groups.GroupedFields := Copy(TempStr, 1, GroupPos - 1);
    Groups.Grouped := Groups.GroupedFields <> '';
    FPageText.Text:=StringReplace(FPageText.Text, '%grouped_fields=' + Groups.GroupedFields + '%', '', [rfIgnoreCase])
  END;
  TempStr := LowerCase(FPageText.Text);
  GroupPos := Pos('%grouped_fields_color=', TempStr);
  IF GroupPos > 0 THEN BEGIN
    Delete(TempStr, 1, GroupPos + Length('%grouped_fields_color=') - 1);
    GroupPos := Pos('%', TempStr);
    Groups.GroupedFieldsColor := Copy(TempStr, 1, GroupPos - 1);
    FPageText.Text:=StringReplace(FPageText.Text, '%grouped_fields_color=' + Groups.GroupedFieldsColor + '%', '', [rfIgnoreCase])
  END;
  TempStr := LowerCase(FPageText.Text);
  GroupPos := Pos('%summary_fields=', TempStr);
  IF GroupPos > 0 THEN BEGIN
    Delete(TempStr, 1, GroupPos + Length('%summary_fields=') - 1);
    GroupPos := Pos('%', TempStr);
    SummaryFields := Copy(TempStr, 1, GroupPos - 1);
    FPageText.Text:=StringReplace(FPageText.Text, '%summary_fields=' + SummaryFields + '%', '', [rfIgnoreCase])
  END;
end;

procedure TEssHtmlReport.SetHtmlText(SText: String);
begin
  FPageText.Text := SText;
  SetColumns;
  SetGroupsFromDocument;
end;

procedure TEssHtmlReport.ParseFieldString(AFields: String; ResultList: TStringList);
var
   StrSepLoc : Integer;
begin
  ResultList.Clear;
  IF AFields = '' THEN Exit;
  Repeat
    StrSepLoc := AnsiPos(';', AFields); // Find the location of the next separator
    IF StrSepLoc = 0 THEN
      ResultList.Add(AFields)
    ELSE BEGIN
      ResultList.Add(Copy(AFields, 1, StrSepLoc - 1));
      Delete(AFields, 1, StrSepLoc);
    END;
    Trim(AFields);
  Until StrSepLoc = 0;
end;

procedure TEssHtmlReport.SetGroups(AGroups: TEssHtmlReportGroups);
begin
  FGroups.Assign(AGroups);
end;

procedure TEssHtmlReport.SetSummaryFields(const ASummaryFields: String);
// SUB FUNCTION
  procedure ParseFieldFunction(ASummaryField: String; var AField, AFunction: String);
  var
     StrSepLoc : Integer;
  begin
     Trim(ASummaryField);
     IF ASummaryFields = '' THEN Exit;
     StrSepLoc := AnsiPos(',', ASummaryField); // Find the location of the next separator
     IF StrSepLoc = 0 THEN
       BEGIN
         AField := LowerCase(ASummaryField);
         Exit;
       END
     ELSE
       AField := LowerCase(Copy(ASummaryField, 1, StrSepLoc - 1));
     Trim(ASummaryField);
     Delete(ASummaryField, 1, StrSepLoc);
     AFunction := ASummaryField;
  end;
// END SUB FUNCTION
var
  AField, AFunction : String;
  i, SummaryIndex : Integer;
begin // Summary field list of format: fieldname,function;fieldname,function;etc...
  FSummaryFieldsList.Clear;
  FSummaryFunctionList.Clear;
  FSummaryFields := ASummaryFields;
  ParseFieldString(FSummaryFields, FSummaryFieldsList);
  FOR i := 0 TO FSummaryFieldsList.Count - 1 DO BEGIN
    ParseFieldFunction(FSummaryFieldsList[i], AField, AFunction);
    IF (AField='') AND (AFunction='') THEN
      Continue
    ELSE IF AField = '' THEN
      ShowMessage(Format('No field defined for function %s', [AFunction]))
    ELSE IF AFunction = '' THEN
      ShowMessage(Format('No function defined for function %s', [AField]))
    ELSE BEGIN
      FSummaryFieldsList[i] := AField;
      FSummaryFunctionList.Add(AFunction);
    END;
  END;
  IF FSummaryFunctionList.Count <> FSummaryFieldsList.Count THEN BEGIN
    FSummaryFieldsList.Clear;
    FSummaryFunctionList.Clear;
  END;
  SummaryIndex := Footer.IndexOf('%add_summary%');
  IF (Footer.Count>0) AND ((FSummaryFieldsList.Count = 0) AND (SummaryIndex >= 0)) THEN
    Footer.Delete(SummaryIndex)
  ELSE IF (FSummaryFieldsList.Count > 0) AND (SummaryIndex < 0) THEN
    Footer.Insert(0, '%add_summary%');  
end;

procedure TEssHtmlReport.SetTableFormat(TableString: String);
var
  TableStopIndex, i : Integer;
  AlignStr : String;
  TableStrings : TStringList;
begin
  TableString := Trim(Copy(TableString, Length('<table') + 1, Length(TableString)));
  TableStopIndex := Pos('>', TableString);
  TableStrings := TStringList.Create;
  TRY
    TableStrings.CommaText := LowerCase(Trim(Copy(TableString, 1, TableStopIndex - 1)));
    FOR i := 0 TO TableStrings.Count - 1 DO
      TableStrings[i] := StringReplace(TableStrings[i], '"', '', [rfReplaceAll]);
    AlignStr := StringReplace(TableStrings.Values['align'], '"', '', [rfReplaceAll]);
    IF AlignStr = 'default' THEN
      TableAttributes.Align := haDefault
    ELSE IF AlignStr = 'left' THEN
      TableAttributes.Align := haLeft
    ELSE IF AlignStr = 'right' THEN
      TableAttributes.Align := haRight
    ELSE IF AlignStr = 'center' THEN
      TableAttributes.Align := haCenter;

    TableAttributes.BGColor := TableStrings.Values['bgcolor'];
    IF TableStrings.Values['border'] = '' THEN
      TableAttributes.Border := -1
    ELSE
      TableAttributes.Border := StrToInt(TableStrings.Values['border']);
    IF TableStrings.Values['cellpadding'] = '' THEN
      TableAttributes.CellPadding := -1
    ELSE
      TableAttributes.CellPadding := StrToInt(TableStrings.Values['cellpadding']);
    IF TableStrings.Values['cellspacing'] = '' THEN
      TableAttributes.CellSpacing := -1
    ELSE
      TableAttributes.CellSpacing := StrToInt(TableStrings.Values['cellspacing']);
    TableStrings.Values['width'] := StringReplace(TableStrings.Values['width'], '%', '', []);
    IF TableStrings.Values['width'] = '' THEN
      TableAttributes.Width := 100
    ELSE
      TableAttributes.Width := StrToInt(StringReplace(TableStrings.Values['width'], '"', '', [rfReplaceAll]));
  FINALLY
    TableStrings.Free;
  END;
end;

procedure TEssHtmlReport.SetColumnFormat(TableString: String);
var
  RowIndex, FieldIndex, StartFieldIndex, EndFieldIndex, AsIndex : Integer;
  ColumnsString, FieldText, AlignStr, VAlignStr : String;
  ColumnValues : TStringList;
  TableColumn : THTMLTableColumn;
begin
  RowIndex := Pos('<tr', TableString);
  IF RowIndex > 0 THEN BEGIN
    ColumnValues := TStringList.Create;
    TRY
      Columns.Clear;
      ColumnsString := Trim(Copy(TableString, RowIndex + 4, Length(TableString)));
      RowIndex := Pos('</tr>', ColumnsString); // Only interested in the first row.
      ColumnsString := Trim(Copy(ColumnsString, 1, RowIndex - 4));
      FieldIndex := Pos('<td', ColumnsString);
      WHILE FieldIndex > 0 DO BEGIN
        ColumnsString := Trim(Copy(ColumnsString, FieldIndex + 4, Length(ColumnsString)));
        ColumnValues.Clear;
        TableColumn := Columns.Add;
        EndFieldIndex := Pos('>', ColumnsString);
        ColumnValues.CommaText := LowerCase(Trim(Copy(ColumnsString, 1, EndFieldIndex - 1)));
        IF ColumnValues.Count > 0 THEN BEGIN
          AlignStr := StringReplace(ColumnValues.Values['align'], '"', '', [rfReplaceAll]);
          IF AlignStr = 'default' THEN
            TableColumn.Align := haDefault
          ELSE IF AlignStr = 'left' THEN
            TableColumn.Align := haLeft
          ELSE IF AlignStr = 'right' THEN
            TableColumn.Align := haRight
          ELSE IF AlignStr = 'center' THEN
            TableColumn.Align := haCenter;
          VAlignStr := StringReplace(ColumnValues.Values['valign'], '"', '', [rfReplaceAll]);
          IF VAlignStr = 'default' THEN
            TableColumn.VAlign := haVDefault
          ELSE IF VAlignStr = 'top' THEN
            TableColumn.VAlign := haTop
          ELSE IF VAlignStr = 'middle' THEN
            TableColumn.VAlign := haMiddle
          ELSE IF VAlignStr = 'bottom' THEN
            TableColumn.VAlign := haBottom
          ELSE IF VAlignStr = 'baseline' THEN
            TableColumn.VAlign := haBaseline;
          TableColumn.BGColor := UpperCase(StringReplace(ColumnValues.Values['bgcolor'], '"', '', [rfReplaceAll]));
        END;
        FieldIndex := Pos('</td>', ColumnsString);
        FieldText := Trim(Copy(ColumnsString, EndFieldIndex + 1, FieldIndex - EndFieldIndex - 1));
        StartFieldIndex := Pos('<', FieldText);
        WHILE StartFieldIndex > 0 DO BEGIN
          EndFieldIndex := Pos('>', FieldText);
          Delete(FieldText, StartFieldIndex, EndFieldIndex);
          StartFieldIndex := Pos('<', FieldText);
        END;
        FieldText := Trim(Copy(FieldText, 1, Length(FieldText)));
        AsIndex := Pos('%as%', FieldText);
        IF AsIndex > 0 THEN
          BEGIN
            TableColumn.FieldName := Trim(Copy(FieldText, 1, AsIndex - 1));
            TableColumn.Title.Caption := Trim(Copy(FieldText, AsIndex + 4, Length(FieldText)));
          END
        ELSE
          TableColumn.FieldName := FieldText;
        FieldIndex := Pos('<td', ColumnsString);
      END;
    FINALLY
      ColumnValues.Free;
    END;
  END;
end;

procedure TEssHtmlReport.SetReportSource(const AFileName: String);
begin
  FPageText.Clear;
  FReportSource := AFileName;
  IF AFileName = '' THEN
    Columns.Clear
  ELSE
    BEGIN
      IF FileExists(AFileName) THEN
        FPageText.LoadFromFile(AFileName)
      ELSE
        ShowMessage('Can not find file : ' + AFileName);
      SetColumns;
      SetGroupsFromDocument;
    END;
end;

function TEssHtmlReport.GetHtmlText: String;
begin
  Result := AdjustLineBreaks(FPageText.Text);
  Result := StringReplace(Result, '#$D', '', [rfReplaceAll, rfIgnoreCase]);
  Result := StringReplace(Result, '#$A', '', [rfReplaceAll, rfIgnoreCase]);
end;

function TEssHtmlReport.GetSummaryFields(APageString: String): String;
var
  GroupedText, SummaryFields, GroupStart, GroupEnd : String;
  StartOfGroup, GroupRow, RowPos:  Integer;
begin
  IF FSummaryFieldsList.Count = 0 THEN BEGIN
    Result := APageString;
    Exit;
  END;
  IF Groups.Grouped THEN
    BEGIN // Summary reports are by group
      StartOfGroup := Pos('%sog=', APageString);
      IF StartOfGroup > 0 THEN BEGIN
        REPEAT
          GroupedText := LowerCase(APageString);
          StartOfGroup := Pos('%sog=', GroupedText);
          Delete(GroupedText, 1, StartOfGroup + 4);

          StartOfGroup := Pos('%sog=', GroupedText);
          IF StartOfGroup = 0 THEN
            BEGIN
              GroupStart := GroupEnd; //
              DataSet.Last;
              GroupEnd := IntToStr(DataSet.RecordCount + 1); //
              Break;
            END
          ELSE
            BEGIN
              GroupStart := Copy(GroupedText, 1, Pos('%', GroupedText) - 1);
              GroupedText := StringReplace(GroupedText, '%sog=' + GroupStart + '%', '', [rfIgnoreCase, rfReplaceAll]);
              StartOfGroup := Pos('%sog=', GroupedText);
              Delete(GroupedText, 1, StartOfGroup + 4);
              GroupEnd := Copy(GroupedText, 1, Pos('%', GroupedText) - 1);
            END;

          APageString := StringReplace(APageString, '%sog=' + GroupStart + '%', '', [rfIgnoreCase, rfReplaceAll]);
          GroupRow := 0;
          RowPos := 0;
          StartOfGroup := Pos('%sog=', APageString);
          GroupedText := LowerCase(APageString);
          WHILE (RowPos + GroupRow < StartOfGroup) DO BEGIN
            RowPos := RowPos + GroupRow;
            GroupRow := Pos('<tr', GroupedText);
            Delete(GroupedText, 1, GroupRow);
            IF GroupRow = 0 THEN Break;
          END;

          FReportFunctions.GroupRecordStart := StrToInt(GroupStart);
          FReportFunctions.GroupRecordEnd := StrToInt(GroupEnd);
          SummaryFields := GetSummaryRow;
          Insert(SummaryFields, APageString, RowPos-1);
        UNTIL StartOfGroup <= 0;

        FReportFunctions.GroupRecordStart := StrToInt(GroupStart);
        FReportFunctions.GroupRecordEnd := StrToInt(GroupEnd);
      END
      ELSE
        FReportFunctions.Grouped := False;
    END;
  SummaryFields := GetSummaryRow;
  SummaryFields := SummaryFields + '</table>';
  APageString := StringReplace(APageString, #$D, '', [rfReplaceAll]);
  APageString := StringReplace(APageString, #$A, '', [rfReplaceAll]);
  Result := StringReplace(APageString, '</table>%add_summary%', SummaryFields, [rfIgnoreCase]);
end;

function TEssHtmlReport.GetSummaryRow: String;
var
  AValue : Variant;
  AFieldName, ColAlign, FunctionResult : String;
  i, FieldIndex : Integer;
begin
  Result := '<tr>';
  FOR i := 0 TO Columns.Count - 1 DO BEGIN
    AFieldName := LowerCase(Columns[i].FieldName);
    FieldIndex := FSummaryFieldsList.IndexOf(AFieldName);
    IF FieldIndex >= 0 THEN
      BEGIN
        CASE Columns[i].Align OF
          haDefault : ColAlign :='default';
          haLeft : ColAlign :='left';
          haCenter : ColAlign :='center';
          haRight : ColAlign :='right';
        END;

        FReportFunctions.EvaluationFunction := FSummaryFunctionList[FieldIndex] + '(' + FSummaryFieldsList[FieldIndex] + ')';
        AValue := FReportFunctions.Evaluate;
        IF (VarType(AValue) = varSmallInt) OR (VarType(AValue) = varInteger) THEN
          FunctionResult := IntToStr(VarAsType(AValue, varInteger))
        ELSE IF (VarType(AValue) = varSingle) OR (VarType(AValue) = varDouble) OR (VarType(AValue) = varCurrency) THEN
          FunctionResult := FloatToStrF(VarAsType(AValue, varDouble), ffFixed, 15, 2)
        ELSE IF (VarType(AValue) = varString) THEN
          FunctionResult := VarAsType(AValue, varString);

        IF Groups.GroupedFieldsColor = '' THEN
          Result := Result + '<td align=' + ColAlign + ' bgcolor="' + Columns[i].BGColor + '" >' + Format('<font face="%s" size="%d">', [FFontFace, FFontSize]) + FSummaryFunctionList[FieldIndex] + '&nbsp;=&nbsp;' + FunctionResult + '</font>'
        ELSE
          Result := Result + '<td align=' + ColAlign + ' bgcolor="' + Groups.GroupedFieldsColor + '" >' + Format('<font face="%s" size="%d">', [FFontFace, FFontSize]) + FSummaryFunctionList[FieldIndex] + '&nbsp;=&nbsp;' + FunctionResult + '</font>'
      END
    ELSE
      BEGIN
        IF FEmptyCellColor='' THEN
          Result := Result + '<td>&nbsp;</td>'
        ELSE
          Result := Result + '<td bgcolor="' + FEmptyCellColor + '">&nbsp;</td>';
      END;
  END;
  Result := Result + '</tr>';
end;

function TEssHtmlReport.GetFunctions(APageString: String): String;
var
  FunctionIndex, EqualPos : Integer;
  AFunction, ReplaceFunction : String;
  AValue : Variant;
begin
  Result := APageString;
  Repeat
    FunctionIndex := Pos('%function', APageString);
    IF FunctionIndex > 0 THEN BEGIN
      ReplaceFunction := Copy(APageString, FunctionIndex + 1, Length(APageString));
      ReplaceFunction := Trim('%' + Copy(ReplaceFunction, 1, Pos('%', ReplaceFunction)));
      EqualPos := Pos('=', ReplaceFunction);
      AFunction := Copy(ReplaceFunction, EqualPos + 1, Length(ReplaceFunction) - EqualPos - 1);
      AFunction := StringReplace(AFunction, ' ', '', [rfReplaceAll]);
      AFunction := StringReplace(AFunction, '<br>', '', [rfReplaceAll]);
      AFunction := StringReplace(AFunction, '<p>', '', [rfReplaceAll]);
      AFunction := StringReplace(AFunction, '</p>', '', [rfReplaceAll]);
      AFunction := StringReplace(AFunction, #$D, '', [rfReplaceAll]);
      AFunction := StringReplace(AFunction, #$A, '', [rfReplaceAll]);
      FReportFunctions.EvaluationFunction := AFunction;
      AValue := FReportFunctions.Evaluate;
      Delete(APageString, 1, FunctionIndex + Length(ReplaceFunction) + 1);
      IF (VarType(AValue) = varSmallInt) OR (VarType(AValue) = varInteger) THEN
        Result := StringReplace(Result, ReplaceFunction, IntToStr(VarAsType(AValue, varInteger)), [])
      ELSE IF (VarType(AValue) = varSingle) OR (VarType(AValue) = varDouble) OR (VarType(AValue) = varCurrency) THEN
        Result := StringReplace(Result, ReplaceFunction, FloatToStrF(VarAsType(AValue, varDouble), ffFixed, 15, 2), [])
      ELSE IF (VarType(AValue) = varString) THEN
        Result := StringReplace(Result, ReplaceFunction, VarAsType(AValue, varString), []);
    END;
  Until FunctionIndex = 0;
end;

function TEssHtmlReport.GetSQL: String;
var
  SqlIndex, EqualIndex : Integer;
  PageString : String;
begin
  IF FPageText = Nil THEN
    ShowMessage('Report source file not assigned.')
  ELSE BEGIN
    PageString := FPageText.Text;
    SqlIndex := Pos('%sql', PageString);
    IF SqlIndex > 0 THEN BEGIN
      Result := Trim(Copy(PageString, SqlIndex + 4, Length(PageString) - SqlIndex));
      EqualIndex := Pos(Result, '=');
      Result := Trim(Copy(Result, EqualIndex + 1, Length(Result) - EqualIndex - 1));
    END;
  END;
end;

procedure TEssHtmlReport.DoFormatCell(CellRow, CellColumn: Integer;
          var BgColor: THTMLBgColor;
          var Align: THTMLAlign; var VAlign: THTMLVAlign;
          var CustomAttrs, CellData: string);
var
  FieldIndex, MatchIndex : Integer;
begin

  IF (CellRow = 0) THEN BEGIN
    IF ShowNiceFields THEN BEGIN
      BgColor := TableAttributes.BgColor;
      CellData := LowerCase(StringReplace(CellData, '_', ' ', []));
      CellData := UpperCase(CellData[1]) + Copy(CellData, 2, Length(CellData));
    END;
  END;

  Inherited DoFormatCell(CellRow, CellColumn, BgColor, Align, VAlign, CustomAttrs, CellData);
  
  IF (CellRow > 0) THEN BEGIN
    IF (CellData = '') AND (FEmptyCellColor<>'') THEN
      CustomAttrs := CustomAttrs + 'BgColor="' + FEmptyCellColor + '"'
    ELSE IF FDataCellColor <> '' THEN
      CustomAttrs := CustomAttrs + 'BgColor="' + FDataCellColor + '"';
  END;

  IF Groups.Grouped AND (CellRow > 0) THEN BEGIN
    IF Groups.FGroupListValues.Count < Groups.FGroupFieldsList.Count THEN
      BEGIN
        FieldIndex := Groups.FGroupFieldsList.IndexOf(Columns[CellColumn].FieldName);
        IF FieldIndex >= 0 THEN BEGIN
          Groups.FGroupListValues.Add(DataSet.FieldByName(Columns[CellColumn].FieldName).AsString);
          CustomAttrs := CustomAttrs + ' %sog=' + IntToStr(DataSet.RecNo) + '%'; // sog = start of group
        END;
      END
    ELSE
      BEGIN
        FieldIndex := Groups.FGroupFieldsList.IndexOf(Columns[CellColumn].FieldName);
        IF FieldIndex >= 0 THEN BEGIN // If field is a group field
          MatchIndex := Groups.FGroupListValues.IndexOf(DataSet.FieldByName(Columns[CellColumn].FieldName).AsString);
          IF MatchIndex < 0 THEN
            BEGIN // If grouped field set has moved to the next group
              Groups.FGroupListValues[FieldIndex] := DataSet.FieldByName(Columns[CellColumn].FieldName).AsString;
              CustomAttrs := CustomAttrs + ' %sog=' + IntToStr(DataSet.RecNo) + '%'; // sog = start of group
            END;
        END;
      END;
  END;
  IF CellData = '' THEN
    CellData := '&nbsp;';
  CellData := Format('<font face="%s" size="%d">%s</font>', [FFontFace, FFontSize, CellData]);
end;

// ******* Function parser ********* //

function TOperand.Value : Variant;
begin
  if ParIndex < 0 then
    result := ActualVal
  else
    result := Parameters[ParIndex]
end;

constructor THtmlReportFunction.Create;
begin
  inherited Create;
  FParamsList := TStringList.Create;
  TempPfxLine := TStringList.Create;
  FFunctions := TStringList.Create;
  FFunctions.Add('not');
end;

destructor THtmlReportFunction.Destroy;
begin
  FParamsList.Free;
  TempPfxLine.Free;
  FFunctions.Free;
  inherited destroy;
end;

procedure THtmlReportFunction.SetFunctions(FunctionList : TStringList);
var
  i : Integer;
begin
  FFunctions.Clear;
  FFunctions.Add('not');
  FOR i := 0 TO FunctionList.Count - 1 DO
    FFunctions.Add(LowerCase(FunctionList[i]));
end;

procedure THtmlReportFunction.OmitComment(var CurLine : String);
var
  i, j : integer;
  s : string;
begin
  IF CurLine = '' THEN Exit;
  s := CurLine;
  i := Pos('{', s);
  j := LastDelimiter('}', s);
  if (0 < i) and (i < j) then
    System.Delete(s, i, j - i + 1);
end;

procedure THtmlReportFunction.ConvertOperators(var CurLine : String);
var
  i : integer;
  // Sub Function
  procedure ChangeOperator(const OldOperator, NewOperator: String; var CurLine: String);
  begin
    i := Pos(OldOperator, CurLine);
    WHILE i > 0 DO BEGIN
      System.Delete(CurLine, i, Length(OldOperator));
      System.Insert(NewOperator, CurLine, i);
      IF (OldOperator = ' and ') OR (OldOperator = ' or ') OR (OldOperator = ' not ') THEN
        CurLine := CurLine + ' )';
      i := Pos(OldOperator, CurLine);
    END;
  end;
  // End Sub Function
begin
  IF CurLine = '' THEN Exit;
  ChangeOperator(' or ', '|( ', CurLine);
  ChangeOperator(' or(', '|(', CurLine);
  ChangeOperator(' and ', '&( ', CurLine);
  ChangeOperator(' and(', '&(', CurLine);
  ChangeOperator(' not ', ' not(', CurLine); // NOT Field1=8 and NOT(Field2=9) becomes NOT(Field1=8 and NOT(Field2=9))
  ChangeOperator(' >= ', '@', CurLine);
  ChangeOperator('>=', '@', CurLine);
  ChangeOperator(' <= ', '%', CurLine);
  ChangeOperator('<=', '%', CurLine);
  ChangeOperator(' <> ', '', CurLine);
  ChangeOperator('<>', '', CurLine);
end;

procedure THtmlReportFunction.myTrim(var s : string);
begin
  //for i := Length(s) downto 1 do
  //  if s[i] <= ' ' then Delete(s, i, 1)
  s := Trim(s);
end;

procedure THtmlReportFunction.SetFilterString(const ValString : String);
var  // ValString is a string of the variables and functions, with each expression separated by a semicolon;
  // e.g.  Var1=5; Var2=6; Y=Var1*Var2; {Two variables and one function}
  LocalStr : string;
begin
  FParamsList.Clear;
  FIsCorrect := False;
  Count := 0;
  FFilterStr := ValString;
  IF ValString = '' THEN EXIT;
  LocalStr := LowerCase(ValString);
  OmitComment(LocalStr);
  ConvertOperators(LocalStr);
  myTrim(LocalStr);
  IF LocalStr <> '' THEN BEGIN
    Count := Count + 1;
    SetLength(Parameters, Count);
    SetLength(PostFixItems, Count);
    ParseLine(LocalStr); // e.g. TempStr : Var1=5
  end;
  FIsCorrect := true;
end;

function THtmlReportFunction.CurSegm : string;
begin
  result := Copy(FWorkStr,1,cur-1);
end;

function THtmlReportFunction.ParseLine(const CurLine: String): Boolean;
begin
  Result := False;
  FParamsList.Add(CurLine);
  FWorkStr := CurLine;
  cur := 1;
  FWorkStrLength := Length(FWorkStr);
  NextToken;
  TempPfxLine.Clear;
  Expression; {input: WorkStr; output: TempPfxLine}
  Init;
  CheckIfFinished; {may raise exceptions}
  Result := True;
end;

procedure THtmlReportFunction.Init;
var
  i : integer;
  Name : string;
begin
  SetLength(PostFixItems[FParamsList.Count-1].Items, TempPfxLine.Count);
  PostFixItems[FParamsList.Count-1].Count := TempPfxLine.Count;
  for i := 0 to PostFixItems[FParamsList.Count-1].Count - 1 do BEGIN
    with PostFixItems[FParamsList.Count-1].Items[i] do begin
      Name := TempPfxLine.Strings[i];
      Operand.ParIndex := -1;
      if Pos(Name, AllowedOp) > 0 then
        op := Ord(Name[1])
      ELSE IF Functions.IndexOf(Name) >= 0 then
        op := Ord(Functions.IndexOf(Name)) + 256
      else
        begin
          op := nop;
          Operand.ActualVal := Name;
        end
    end;
  END;
end;

procedure THtmlReportFunction.NextToken;
begin
  FToken := '';
  if cur > FWorkStrLength then Exit;
  if FWorkStr[cur] in AlfaNum then {returns a character name or number}
    while (cur <= FWorkStrLength) and (FWorkStr[cur] in AlfaNum) do begin
      FToken := FToken + FWorkStr[cur];
      Inc(cur);
    end
  else {returns an operation sign}
    begin
      FToken := FWorkStr[cur];
      Inc(cur)
    end;
  FToken := Trim(FToken)
end;

procedure THtmlReportFunction.Expression;
var
  TermOp, FactOp : char;
  i : shortint;
  FunctionName : String;

  // **** SUB PROCEDURE ****//
  procedure Term;

    // **** SUB PROCEDURE ****//
    procedure Factor;
    begin
      i := Functions.IndexOf(FToken);
      if i >= 0 then {correct function name}
        begin
          FunctionName := FToken;
          NextToken;
          if FToken <> '(' then
            raise Exception.CreateFmt('%s'#10#13' "(" expected but "%s" found', [CurSegm,FToken]);
        end;
      if FToken = '(' then {factor is a function or expression}
        begin
          NextToken;
          Expression; {recursive!}
          if FToken <> ')' then
            raise Exception.CreateFmt('%s'#10#13' ")" expected but "%s" found', [CurSegm,FToken]);
          NextToken;
          if i >= 0 then {function #i}
            begin
              TempPfxLine.Add('0'); {2nd operand 0}
              TempPfxLine.Add(FunctionName); // FunctionName will be passed to the OnEvaluateFunction method by its integer reference in the Functions string List.
            end
        end {factor is a function or expression}
      else {factor is a variable or number}
        begin
          if FToken = '' then
            raise Exception.CreateFmt('%s'#10#13' operand expected but nothing found', [CurSegm]);
          TempPfxLine.Add(FToken);
          NextToken;
        end; {factor is a variable or number, possibly followed by power index}

      while FToken = '^' do begin
        NextToken;
        Factor; {recursive!}
        TempPfxLine.Add('^')
      end
    end;
    // **** END fACTOR SUB PROCEDURE ****//

  begin {Term}
    Factor;
    while (FToken = '*') or (FToken = '/') or (FToken = '>') or (FToken = '<') or (FToken = '=') or (FToken = '&') or (FToken = '|') or (FToken = '%') or (FToken = '@') or (FToken = '') do begin
      FactOp := FToken[1];
      NextToken;
      Factor;
      TempPfxLine.Add(FactOp);
    end
  end;
  // **** END TERM SUB PROCEDURE ****//
begin {Expression}
  if (FToken = '+') or (FToken = '-') then
    TempPfxLine.Add('0') {0 +... or 0 -...}
  else
    Term;
  while (FToken = '+') or (FToken = '-') do begin
    TermOp := FToken[1];
    NextToken;
    Term;
    TempPfxLine.Add(TermOp)
  end;
end; {Expression}

procedure THtmlReportFunction.CheckIfFinished;
begin
  if cur <= FWorkStrLength then
    raise Exception.CreateFmt('%s'#10#13' parsing stopped here', [CurSegm]);
  if FToken <> '' then
    raise Exception.CreateFmt('%s'#10#13' can not be parsed', [FToken])
end;

function THtmlReportFunction.Evaluate : Variant;
var
  i : integer;
begin
  FFilterBoolean := True;
  if not IsCorrect then
    raise Exception.Create('The formula list has an error.');
  for i := 0 to Count - 1 do
    Parameters[i] := PostFixEvaluate(i);
  result := Parameters[Count - 1];
end;

function THtmlReportFunction.PostfixEvaluate(i: Integer) : variant;
var
  Pos : cardinal; {must point the position next to desired postfix start}
  // **** SUB FUNCTION **** //
  function GetResult(APostFix: THtmlReportPostfix) : Variant;
  begin
    Dec(Pos); {Set Range Check ON: Pos < 0 means a mistake in postfix sequence}
    with APostFix.Items[Pos] do
      if op = nop then
        result := Operand.Value
      else
        result := Self.FilterOperations(op, GetResult(APostFix), GetResult(APostFix));
  end; {Left-to-Right parameter evaluation essential}
  // **** END GetResult SUB FUNCTION **** //
begin
  Pos := PostFixItems[i].Count;
  result := GetResult(PostFixItems[i]);
end;

function THtmlReportFunction.FilterOperations(op: Word; y: variant; x: Variant) : Variant;
var
  n : integer;
  AFunctionName : String;
begin
  TRY
  case Chr(op) of // max function op list of 41 since ORD('*') = 42
    '@' : BEGIN
            IF FDataSet.FieldByName(Trim(y)).DataType = ftString THEN
              Result := Boolean(AnsiCompareText(FDataSet.FieldByName(Trim(y)).Value, X) >= 0)
            ELSE
              Result := (FDataSet.FieldByName(Trim(y)).Value >= X); // two character function of >=
          END;
    '%' : BEGIN
            IF FDataSet.FieldByName(Trim(y)).DataType = ftString THEN
              Result := Boolean(AnsiCompareText(FDataSet.FieldByName(Trim(y)).Value, X) <= 0)
            ELSE
              Result := (FDataSet.FieldByName(Trim(y)).Value <= X); // two character function of <=
          END;
    '' : BEGIN
            IF FDataSet.FieldByName(Trim(y)).DataType = ftString THEN
              Result := Boolean(AnsiCompareText(FDataSet.FieldByName(Trim(y)).Value, X) <> 0)
            ELSE
              Result := (FDataSet.FieldByName(Trim(y)).Value <> X); // two character function of <>
          END;
    '*' : Result := x * y;
    '+' : Result := x + y;
    '-' : Result := y - x;
    '/' : TRY
            Result := y / x;
          EXCEPT
            ON EMathError DO Result := 'NAN';
          END;
    '^' : BEGIN
            if Frac(y) <> 0.0 then
              result := Power(x, y)
            else
              begin
                n := Floor(y);
                result := IntPower(x, n);
              end;
          END;
    '=' : BEGIN
            IF FDataSet.FieldByName(Trim(y)).DataType = ftString THEN
              Result := Boolean(AnsiCompareText(FDataSet.FieldByName(Trim(y)).Value, X) = 0)
            ELSE
              Result := (FDataSet.FieldByName(Trim(y)).Value = X);
          END;
    '<' : BEGIN
            IF FDataSet.FieldByName(Trim(y)).DataType = ftString THEN
              Result := Boolean(AnsiCompareText(FDataSet.FieldByName(Trim(y)).Value, X) < 0)
            ELSE
              Result := (FDataSet.FieldByName(Trim(y)).Value < X);
          END;
    '>' : BEGIN
            IF FDataSet.FieldByName(Trim(y)).DataType = ftString THEN
              Result := Boolean(AnsiCompareText(FDataSet.FieldByName(Trim(y)).Value, X) > 0)
            ELSE
              Result := (FDataSet.FieldByName(Trim(y)).Value > X);
          END;
    '&' : Result := (Y AND X); //FFilterBoolean AND x;
    '|' : Result := (Y OR X); //FFilterBoolean OR x;
    ELSE BEGIN// all function calls are handled externally by the calling program depending on the list of Functions avaiable
      AFunctionName := Functions[Ord(op) - 256];
      IF AFunctionName = 'not' THEN
        BEGIN
          FFilterBoolean := NOT y;
          Result := VarAsType(FFilterBoolean, varBoolean);
        END
      ELSE IF AFunctionName = 'count' THEN
        BEGIN
          FDataSet.Last;
          IF Grouped THEN
            Result := GroupRecordEnd -  GroupRecordStart
          ELSE
            Result := FDataSet.RecordCount;
        END
      ELSE IF AFunctionName = 'sum' THEN
        BEGIN
          FDataSet.First;
          Result := 0;
          WHILE NOT FDataset.EOF DO BEGIN
            IF Grouped THEN
              BEGIN
                IF (FDataSet.RecNo >= GroupRecordStart) AND (FDataSet.RecNo < GroupRecordEnd) THEN
                  Result := Result + FDataSet.FieldByName(y).Value;
              END
            ELSE
              Result := Result + FDataSet.FieldByName(y).Value;
            FDataSet.Next;
          END;
        END
      ELSE IF AFunctionName = 'avg' THEN
        BEGIN
          FDataSet.First;
          Result := 0;
          WHILE NOT FDataset.EOF DO BEGIN
            IF Grouped THEN
              BEGIN
                IF (FDataSet.RecNo >= GroupRecordStart) AND (FDataSet.RecNo < GroupRecordEnd) THEN
                  Result := Result + FDataSet.FieldByName(y).Value;
              END
            ELSE
              Result := Result + FDataSet.FieldByName(y).Value;
            FDataSet.Next;
          END;
          TRY
            Result := Result / FDataSet.RecordCount;
          EXCEPT
            ON EMathError DO Result := 'NAN';
          END;
        END
      ELSE IF AFunctionName = 'min' THEN
        BEGIN
          FDataSet.First;
          Result := FDataSet.FieldByName(y).Value;
          WHILE NOT FDataset.EOF DO BEGIN
            IF Grouped THEN
              BEGIN
                IF (FDataSet.RecNo >= GroupRecordStart) AND (FDataSet.RecNo < GroupRecordEnd) THEN
                  IF Result > FDataSet.FieldByName(y).Value THEN
                    Result := FDataSet.FieldByName(y).Value;
              END
            ELSE
              BEGIN
                IF Result > FDataSet.FieldByName(y).Value THEN
                  Result := FDataSet.FieldByName(y).Value;
              END;
            FDataSet.Next;
          END;
        END
      ELSE IF AFunctionName = 'max' THEN
        BEGIN
          FDataSet.First;
          Result := 0;
          WHILE NOT FDataset.EOF DO BEGIN
            IF Grouped THEN
              BEGIN
                IF (FDataSet.RecNo >= GroupRecordStart) AND (FDataSet.RecNo < GroupRecordEnd) THEN
                  IF Result < FDataSet.FieldByName(y).Value THEN
                    Result := FDataSet.FieldByName(y).Value;
              END
            ELSE
              BEGIN
                IF Result < FDataSet.FieldByName(y).Value THEN
                  Result := FDataSet.FieldByName(y).Value;
              END;
            FDataSet.Next;
          END;
        END
      ELSE IF Assigned(OnEvaluateFunction) THEN // Handle all unassigned functions here
        OnEvaluateFunction(x, y, Functions[Ord(op) - 256], Result);
    END;
  end
  EXCEPT
    Result := 'NAN';
  END;
end;

end.

