unit THtmlUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, DBTables, OleCtrls, SHDocVw, HTTPApp, DBWeb, ESSHtmlReport, StdCtrls,
  DBCtrls, ExtCtrls, Htmlview, Buttons, jpeg;

type
  TForm1 = class(TForm)
    ESSHtmlReport1: TESSHtmlReport;
    Table1: TTable;
    Button1: TButton;
    HTMLViewer: THTMLViewer;
    Label1: TLabel;
    Label2: TLabel;
    BitBtnPrintPreview: TBitBtn;
    Label3: TLabel;
    Label4: TLabel;
    ImageMoon: TImage;
    procedure Button1Click(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ESSHtmlReport1FormatCell(Sender: TObject; CellRow,
      CellColumn: Integer; var BgColor: THTMLBgColor;
      var Align: THTMLAlign; var VAlign: THTMLVAlign; var CustomAttrs,
      CellData: String);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure HTMLViewerImageRequest(Sender: TObject; const SRC: String;
      var Stream: TMemoryStream);
    procedure BitBtnPrintPreviewClick(Sender: TObject);
  private
    { Private declarations }
    FImageList : TStringList;
    procedure ClearList(AList: TStringList);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

uses PreviewForm;  // Preview form is in the Demo_Src folder where you installed Thtml.  Add this to your project search path

procedure TForm1.Button1Click(Sender: TObject);
var
  AHtmlFile : OleVariant;
  MoonImage : TMemoryStream;
begin
  //  ESSHtmlReport1.ReportSource := some file on disk
  ClearList(FImageList);
  // We have to add our header logo to the list of pictures
  MoonImage := TMemoryStream.Create;
  ImageMoon.Picture.Graphic.SaveToStream(MoonImage);
  FImageList.AddObject('moon.jpg', MoonImage);

  AHtmlFile := ExtractFilePath(Application.ExeName) + 'temp.htm';
  Table1.Open;
  ESSHtmlReport1.Content;
  HtmlViewer.LoadStrings(EssHtmlReport1.PageText);
  Table1.Close;
  BitBtnPrintPreview.Enabled := True;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  //WebBrowser1.Height := Self.Height - 75;
end;

procedure TForm1.ESSHtmlReport1FormatCell(Sender: TObject; CellRow,
  CellColumn: Integer; var BgColor: THTMLBgColor; var Align: THTMLAlign;
  var VAlign: THTMLVAlign; var CustomAttrs, CellData: String);
var
  AFileName, DataString : String;
  AStream : TStream;
  AStream2 : TMemoryStream;
begin
  IF CellRow = 0 THEN Exit;
  // Use THtml component set (from: http://www.pbear.com) and its OnImageRequest
  // event if you do not want to create files of the image.
  IF EssHtmlReport1.DataSet.Fields[CellColumn].DataType = ftTypedBinary THEN BEGIN
    AFileName := EssHtmlReport1.DataSet.FieldByName('NAME').AsString;
    IF Not FileExists(AFileName) THEN BEGIN
      AStream := EssHtmlReport1.DataSet.CreateBlobStream(EssHtmlReport1.DataSet.Fields[CellColumn], bmRead);
      AStream2 := TMemoryStream.Create;
      TRY
        AStream.Seek(8, soFromBeginning);  // Eliminate the DBase graphic header bits.
        AStream2.CopyFrom(AStream, AStream.Size - 8);
        FImageList.AddObject(AFileName, AStream2);   // add the picture to the list to be called later by the Html.OnImageRequest method
      Finally
        AStream.free;
      End;
    END;
    CellData := '<img src="' + AFileName + '" width="130" height="88">';
  END;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  FImageList := TStringList.Create;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  ClearList(FImageList);
end;

procedure TForm1.ClearList(AList: TStringList);
var
  i : Integer;
begin
  FOR i := 0 TO AList.Count - 1 DO
    AList.Objects[i].Free;
end;

procedure TForm1.HTMLViewerImageRequest(Sender: TObject;
  const SRC: String; var Stream: TMemoryStream);
var
  ImageIndex : Integer;
begin
  ImageIndex := FImageList.IndexOf(SRC);
  IF ImageIndex < 0 THEN Exit;
  Stream := TMemoryStream(FImageList.Objects[ImageIndex]);
end;

procedure TForm1.BitBtnPrintPreviewClick(Sender: TObject);
var
  pf: TPreviewForm;     // a THtml component form.  In ...\Thtml\Demo_Src\PreviewForm.pas
  Abort: boolean;
begin
  Application.ProcessMessages;
  pf := TPreviewForm.CreateIt(Self, HTMLViewer, Abort);
  TRY
    IF NOT Abort THEN
      pf.ShowModal;
  FINALLY
    pf.Free;
    Application.ProcessMessages;
  END;
end;

end.
